/*
Copyright (c) 2009, Joel Reardon
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the University of Waterloo nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/



#include "spe.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h> 
#include <pthread.h>
#include <strings.h>
#include <string.h>


int atoi_or_null(char* c) {
	if (c) return atoi(c);
	return 0;
}

int strpre(char* str, char* prefix) {
	int ret;
	while (*prefix && *str) {
		if (*prefix != *str) return 0;
		++prefix;
		++str;	
	}
	ret = !*prefix;
	return ret;
}

void* spe_str_to_ptr(char* str) {
	return (void*)strtoul(str, NULL ,16);
	
}


void spe_parameters_split(char* s, char delim, char** ret, int n) {
	int i = 0;
	int j = 0;
	int state = 0;
	while (s[i] == delim) ++s;	
	ret[j++] = s;
	
	
	while (s[i] && j < n) {
		if (s[i] == delim) {
			s[i] = 0;
			state = 1;
		}
		else {
			if (state == 1) {
				state = 0;
				ret[j++] = s+i;
					  
			}
		}
		++i;
	}
	for (; j < n; ++j) 
		ret[j]=NULL;
	
	while (s[i]) {
		if (s[i] == delim)
			s[i] = 0;
		++i;	
	}
}

void spe_listener_command_dump(spe_t * spe, char* cmd, GString* ret) {
	FILE * fdump = fopen(cmd, "w");
	if (!fdump) {
		g_string_append(ret,"<error text=\"Opening File\"/>");
		return;		
	}
	spe_observer_dump(spe, fdump);
//	spe_buffer_dump(spe, fdump);
	spe_variable_dump(spe, fdump);

	fclose(fdump);
}

void spe_listener_command_list(spe_t * spe, char* cmd, GString* ret) { 
	if (strpre(cmd, "ALL")) {
		spe_observer_list(spe, ret);
//		spe_buffer_list(spe, ret);
		spe_variable_list(spe, ret);
	}
	else if (strpre(cmd, "OBSERVER")) {
		spe_observer_list(spe, ret);
	}
//	else if (strpre(cmd, "BUFFER")) {
//		spe_buffer_list(spe, ret);
//	}
	else if (strpre(cmd, "VARIABLE")) {
		spe_variable_list(spe, ret);
	}
}

void spe_listener_command_view(spe_t * spe, char* cmd, GString* ret) { 
	char* parms[2];
	spe_parameters_split(cmd, ' ', parms, 2);
	if(!strcmp(parms[0], "OBSERVER")) {
		spe_observer_view(spe, ret, spe_str_to_ptr(parms[1]));		
	}
//	else if(!strcmp(parms[0], "BUFFER")) {
//		spe_buffer_view(spe, ret, spe_str_to_ptr(parms[1]));	
//	}
	else if(!strcmp(parms[0], "VARIABLE")) {
		spe_variable_view(spe, ret, spe_str_to_ptr(parms[1]));
		
	}
}


void trace_parms(char** parms, int n) {
	int i;
	for (i = 0 ; i < n ; ++i) {
		if (parms[i])
			printf("parms[%d] = %s\n", i,parms[i]);
	}
}
void spe_listener_command_seta(spe_t * spe,char* cmd, GString* ret) { 
	int newstate;
	int fam;
	int quota;
	struct timeval interval;
	char* parms[6];
	char* filename;
	int state;
	
	spe_parameters_split(cmd, ' ', parms, 6);
	
	fam = atoi_or_null(parms[0]);
	state = atoi_or_null(parms[1]);
	quota = atoi_or_null(parms[2]);
	interval.tv_sec = atoi_or_null(parms[3]);
	interval.tv_usec = atoi_or_null(parms[4]);

	filename = parms[5]; //might be NULL if missing parameter
	spe_observer_set_state_all(spe, fam, state, quota, &interval, filename);

}

void spe_listener_command_seto(spe_t * spe,char* cmd, GString* ret) { 
	int newstate;
	void* ptr;
	int quota;
	struct timeval interval;
	char* parms[6];
	char* filename;
	int state;
	
	spe_parameters_split(cmd, ' ', parms, 6);
	
	ptr = spe_str_to_ptr(parms[0]);
	state = atoi_or_null(parms[1]);
	quota = atoi_or_null(parms[2]);
	interval.tv_sec = atoi_or_null(parms[3]);
	interval.tv_usec = atoi_or_null(parms[4]);
	if (interval.tv_sec == 0 && interval.tv_usec == 0) interval.tv_sec = 100000;
	filename = parms[5]; //might be NULL if missing parameter
	spe_observer_set_state(spe, ptr, state, quota, &interval, filename);

}

/*
 * void spe_listener_command_setb(spe_t * spe,char* cmd, GString* ret) { 
	int newstate;
	void * ptr;

	newstate  = cmd[0];
	++cmd;
	ptr = spe_str_to_ptr(cmd);
	spe_buffer_set_state(spe, ptr, newstate);

}
* */

void spe_listener_command_setv(spe_t * spe,char* cmd, GString* ret) { 
	int newstate;
	void * ptr;

	newstate  = cmd[0];
	++cmd;
	ptr = spe_str_to_ptr(cmd);
	spe_variable_set_state(spe, ptr, newstate);

}
void spe_listener_command_hogm(spe_t* spe, char *cmd, GString* ret) {
	int toggle = atoi_or_null(cmd);
	spe_variable_stat(spe, toggle);
	
}

void spe_listener_command_flib(spe_t* spe, char *cmd, GString* ret) {
	
	g_string_append_printf(ret, "<error text=\"%s\">", spe_fl_load(spe->obs_mgr->fl, cmd));
	
}
void spe_listener_command_begin(spe_t* spe, char* cmd, GString* ret) {

	spe_observer_begin(spe, cmd);


}
void spe_listener_command_end(spe_t* spe, char* cmd, GString* ret) {

	spe_observer_end(spe);


}


void * spe_listener_interaction_thread (void*  param) {
	char buffer[1024];
	
	int n;

	gchar* reply;
	GString* send_string = g_string_new("");
	
	void** params = param;
	spe_t* spe= params[0];
	int socket = (int)params[1];
	free(params);
	if (socket< 0) {
		fprintf(stderr,"ERROR on accept");
		return NULL;
	}
	
	bzero(buffer,1024);
	n = recv(socket,buffer,1023, 0);

	printf("recv: %s\n", buffer);
	if (n<=0) return NULL;
	
	if (strpre(buffer, "DUMP")) {
		spe_listener_command_dump(spe, buffer+5, send_string);
	}
	else if (strpre(buffer, "VIEW")) {
		spe_listener_command_view(spe, buffer+5, send_string);
	}
	else if (strpre(buffer, "LIST")) {
		spe_listener_command_list(spe, buffer+5, send_string);
	}
	else if (strpre(buffer, "SETA")) {
			//set all
		spe_listener_command_seta(spe, buffer+5, send_string);			
	}
	else if (strpre(buffer, "SETO")) {
		spe_listener_command_seto(spe, buffer+5, send_string);
	}
//	else if (strpre(buffer, "SETB")) {
//		spe_listener_command_setb(spe, buffer+5, send_string);
//	}
	else if (strpre(buffer, "SETV")) {
		spe_listener_command_setv(spe, buffer+5, send_string);
	}
	else if (strpre(buffer, "FLIB")) {
		spe_listener_command_flib(spe, buffer+5, send_string);
	}
	else if (strpre(buffer, "HOGM")) {
		spe_listener_command_hogm(spe, buffer+5, send_string);
	}
	else if (strpre(buffer, "CORE DUMP")) {
		abort();
	}
	else if (strpre(buffer, "ECHO")) {
		g_string_append_printf(send_string,"ECHO\n");
	}
	else if (strpre(buffer, "BEGIN")) {
		spe_listener_command_begin(spe, buffer+6, send_string);
	}
	else if (strpre(buffer, "END")) {
		spe_listener_command_end(spe, buffer+4, send_string);
	}

	else if (strpre(buffer, "EXIT NOW")) {
		spe->alive = 0;
	}
	else if (strpre(buffer, "FCLOSEALL")) {
		fcloseall();
	}
	
	
  
	reply = g_string_free(send_string, FALSE);
	
	n = send(socket,reply,strlen(reply), 0);
	free(reply);
	shutdown(socket, SHUT_RDWR);
	
	pthread_exit(NULL);
}

void* spe_listener_thread(spe_t * spe) {
	
	int newsockfd, portno;
	socklen_t  clilen;
	struct sockaddr_in serv_addr, cli_addr;
    portno = spe->socket_port;
	spe->socket = socket(AF_INET, SOCK_STREAM, 0);
	if (spe->socket < 0) 
		pthread_exit((void*)SPE_SOCKET_ERROR_SOCKET);
	bzero((char *) &serv_addr, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = INADDR_ANY;
	serv_addr.sin_port = htons(spe->socket_port);
	if (bind(spe->socket , (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) 
		pthread_exit( (void*)SPE_SOCKET_ERROR_BIND);
	listen(spe->socket,5);
	clilen = sizeof(cli_addr);
	
	
	while (spe->socket_thread_keepalive) { 
		fd_set fds;
		struct timeval tv;
		int retval;
        tv.tv_usec = 0 ;
		tv.tv_sec = 1;
		
        FD_ZERO(&fds);
        FD_SET(spe->socket, &fds);
        //select() expires every second so that the thread can exit when keepalive is false
		retval = select(1+spe->socket, &fds, NULL, NULL, &tv);
		
		if (retval) {
			newsockfd = accept(spe->socket, (struct sockaddr *) &cli_addr, &clilen);
			if (16777343 == cli_addr.sin_addr.s_addr) {
					// we have a local connection
				pthread_t *newthread = malloc(sizeof(pthread_t));
				void** param = malloc(2*sizeof(void*));
				param[0]= spe;
				param[1]= (void*)newsockfd;
				pthread_create(newthread, NULL, spe_listener_interaction_thread, (void*)param);
				int i;
				pthread_join(*newthread, (void*)&i);
				free(newthread);
				
			}		
			close(newsockfd);
		}
	}
	shutdown(spe->socket,SHUT_RDWR);
	pthread_exit((void*) SPE_SOCKET_SUCCESS); 
}
